/*********************************************************************
 *
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2011, Robert Bosch LLC.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Robert Bosch nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 *********************************************************************/

//\Author Philip Roan and Joerg Wagner, Bosch LLC


#ifndef PROXDRIVER_H_
#define PROXDRIVER_H_

//#include <stdint.h>
#include <libsub.h> //sub20_driver

#include "proximity_sensor_driver/prox_sensor_measurement.h" //message
#include "proximity_sensor_driver/reinitialize_proximity_sensor.h" //service

//dynamic reconfigure
#include <dynamic_reconfigure/server.h>
#include <proximity_sensor_driver/ProximitySensorConfig.h>
                      
/*!
 * \brief Communitcates with the Bosch sensor skin.
 *
 * The class Skin handles the communication with the sensor skin and the GUIs.
 * It reads the sensor values as well as the log files from the sensor skin and
 * publishes them on the topic "skin_data". Apart from that the class also
 * writes data in the ECU of the skin when a service request is sent.
 */
class Skin
{
 public:
  /*!
   *
   * The constructor of the scans all connected USB devices and looks for a
   * Sub20 device with connection to a sensor skin. If a Sub20 device with a
   * connection to a sensor skin is found the number of connected sensor
   * elements will be determined. To determine the number of sensor elements the
   * constructor reads the value "Limit" of each sensor element until the end of
   * chain marker (Limit == 254) is found. This marker is an indicator for the
   * first unused sensor element address.
   */
  Skin(const ros::NodeHandle& nh);
  
  /*!
   * \brief Reset values and to close the opened Sub20 device
   */
  ~Skin();
    
  /*!
   * Get the most recent measurement from all skin cells
   *
   * The function "GetOneMeas" reads the data of the connected sensor elements as well as the log files
   * from the RAM of the sensor skin.
   *
   * To read data from the RAM of the sensor skin a special communication pattern has to be used.
   *
   * Communication pattern:
   * 		1.) Memory address write
   * 			Start|Slave Adr. (Bit1-7) + Bit0 = 0|Mem. Addr. MSB|Mem. Addr. LSB|Stop
   *
   * 		2.) Wait for 50ms
   *
   * 		3.) Data read
   * 			Start|Slave Adr. (Bit1-7) + Bit0 = 1|Data[0]|Data[1]|... |Stop
   */
  bool getMeasurement( void );
  
  void checkConnection( void );
  
  //! Dynamic reconfigure callback
  void reconfigure_callback( proximity_sensor_driver::ProximitySensorConfig &config, uint32_t level );
  bool reinitialize_callback( proximity_sensor_driver::reinitialize_proximity_sensor::Request &req,
			      proximity_sensor_driver::reinitialize_proximity_sensor::Response &res );
  
  // private:

  enum skin_cell_shape { SQUARE, RECTANGLE, TRIANGLE };
  struct skin_cell
  {
    uint8_t status;
    uint8_t value;
    uint8_t offset;
    uint8_t limit;
    uint8_t U1;
    uint8_t U2;
    uint8_t U3;
    uint8_t dynamic;
  };
  enum ECU_command { INITIALIZE, WRITE_TO_FLASH, SET_DYNAMIC_THRESHOLD, SET_STATIC_THRESHOLD, SET_END_OF_CHAIN, ACTIVATE_SENSOR, DEACTIVATE_SENSOR };
  enum sensor_activity { ACTIVE = (char)1, INACTIVE = (char)0 };

  
  //Define I2C communication parameters
  //! Address of slave (sensor skin)
  static const int SKIN_I2C_ADR = 0x01;
  
  static const int RAM_ADR_LENGTH = 2;
  static const int RAM_WRITE_LENGTH = 1;
  //! Give this value for 'ma_sz' in sub_i2c_read/sub_i2c_write to skip the "Memory Address Write" stage
  static const uint8_t SKIP_MEM_ADR_WRITE = 0;
 
  //Define Skin Hardware parameters
  //! Address for command bytes sent to ECU RAM
  static const uint16_t COMMAND_ADR = 0x00C0;
  //! Address of the dynamic threshold in the ECU RAM
  static const uint16_t DYNAMIC_THRESHOLD_ADR = 0x00C2;
  //! Address of the static threshold of the first cell in the ECU RAM
  static const uint16_t STATIC_THRESHOLD_ADR = 0x0103;
  //! Address of of the bit for activating the first sensor.
  static const uint16_t ACTIVATE_ADR = 0x0100;
  //! High byte of the beginning address of the sensor data in ECU RAM
  static const uint8_t DATA_ADR_HI = 0x01;
  //! Low byte of the beginning address of the sensor data in ECU RAM
  static const uint8_t DATA_ADR_LO = 0x00;
  //! High byte of the beginning address of the status data in ECU RAM
  static const uint8_t STATUS_ADDRESS_HIGH = 0x00;
  //! Low byte of the beginning address of the status data in ECU RAM
  static const uint8_t STATUS_ADDRESS_LOW  = 0xC1;
  //! Length of status message in bytes
  static const uint8_t STATUS_LENGTH = 2;

  
  //! Maximum number of connectable sensors
  static const int MAX_SENSORS = 192;
  //! Maximum number of bytes which can be read over a single I2C transmission
  static const int MAX_BYTES_PER_I2C = 256;
  //! The number of sensors which can be queried in a single I2C transmission
  static const int SENSORS_PER_I2C = MAX_BYTES_PER_I2C / sizeof(struct skin_cell);
  //! Value of the end of chain marker
  static const uint8_t END_OF_CHAIN_MARKER = 254;
  //! Default value for thresholds if not set by parameters in yaml file
  static const uint8_t THRESHOLD_DEFAULT = 10;

  
  // skin class members for Xdimax Sub20-Device
  
  //! device handle
  sub_handle handle_;
  sub_device subdev_;
  //! skin connection state
  bool skin_connected_;
  // device serial number
  std::string serialNumber_;
  // I2C communication frequency (standard mode: 100 kHz)
  sub_int32_t i2c_frequency_; // value set in constructor
 

  // skin class members for skin cells

  //! Array containing measurements for each cell
  struct skin_cell cells_[MAX_SENSORS];
  //! Array containing shape of each cell
  skin_cell_shape shapes_[MAX_SENSORS];
  //! Number of sensor elements connected to the ECU of the sensor skin
  int cell_count_;
  //! Noise threshold in sensor readings
  uint8_t noise_;
  //! Obstacle detection threshold
  uint8_t threshold_;
  //! Value of dynamic threshold
  uint8_t dynamic_threshold_;

  //! skin initialization state
  // bool binit_skin_ok_;
  //! This flag is true if a valid measurement is available
  // bool bMeasAvailable_;
  
  // Status of skin (Go/NoGo - Bit0, Err/NoErr - Bit1)
  bool status_go_;
  bool status_err_;

  bool init( void );

  bool sendToECU( ECU_command command, uint8_t sensor_number, uint8_t data);
  
  /*!
   * \brief Converts a char array into an uint64_t array
   *
   * The function "convert_data_2_uint64_array" receives a char array of size 1536 and returns an
   * uint64 array of size 192.
   *
   * \param inp_buf_data Received char array
   * \param outp_sen_data Returned uint64 array
   */
  //void convert_data_2_uint64_array(char* inp_buf_data, uint64_t* outp_sen_data);

};//class


#endif /* PROXDRIVER_H_ */
